import os
import io
import re
import json
import shutil
import tempfile
import itertools
import zipfile
import numpy as np

from typing import Any
from typing import Dict
from typing import Generator
from typing import Optional
from typing import Tuple
from typing import Generator
from typing import Tuple
from typing import TypeVar


import pytpu as tpu


_TPU_PROGRAM_MEMBER_NAME = 'program.tpu'
_INPUTS_MEMBER_PREFIX = 'inputs'
_OUTPUTS_MEMBER_PREFIX = 'outputs'
_META_DATA_MEMBER_NAME = 'metadata.json'


MetaData = Dict  # Untyped JSON meda data dict


absolute_threshold = relative_threshold = 1e-2



def get_data_path():
    return os.path.join(
        os.path.dirname(os.path.realpath(__file__)),
        "../data"
    )


def get_programs_path(tpu_device: tpu.Device):
    dev_info = tpu_device.info()

    ddr_word_length = dev_info["axi_word_len"] // 8
    cache_word_length = dev_info["cache_word_len"] // 8
    programs_path = os.path.join(
        get_data_path(),
        "ddr_{}/cache_{}".format(ddr_word_length, cache_word_length)
    )
    if not os.path.isdir(programs_path):
        raise RuntimeError(
            "No device pytpu_tests for axi word length {} cache word length {}".format(ddr_word_length,
                                                                                       cache_word_length))
    return programs_path

def _sanitize_layer_name(name: Any) -> str:
    name = str(name)
    name = re.sub(r'[\,|,/,?,<,>,+,",*,:, ,.,%,!,@,(,),[,\]]+', r'_', name)
    return name


def _read_meta_data(tpu_program_zip: zipfile.ZipFile) -> MetaData:
    return json.load(io.BytesIO(tpu_program_zip.read(_META_DATA_MEMBER_NAME)))


def _read_tensor_dict(
        tpu_test: zipfile.ZipFile,
        meta_data: Dict,
        zip_file_member_path_prefix: str,
) -> Generator[Tuple[str, np.ndarray], None, None]:
    for _, region in meta_data[zip_file_member_path_prefix].items():
        for tensor_name, tensor_params in region.items():
            file_name = zip_file_member_path_prefix + '/' + _sanitize_layer_name(tensor_name) + '.bin'
            tensor_buffer = tpu_test.read(file_name)

            # non raw
            if tensor_params.get("user_dtype") and tensor_params.get("user_shape"):
                tensor_data = np.frombuffer(tensor_buffer, dtype=tensor_params['user_dtype'])
                tensor_data = tensor_data.reshape(tensor_params['user_shape'])
            # raw
            else:
                tensor_data = np.frombuffer(tensor_buffer, dtype="uint8")
                tensor_data = tensor_data.reshape(-1, meta_data['hardware_parameters']['ddr_word_len'])

            # LOGGER.debug(f'External file "{file_name}" associated with tensor "{tensor_name}" is loaded:'
            #              f' {tensor_data.shape} <{tensor_data.dtype}>')
            yield tensor_name, tensor_data


def _inspect_input_zip(
        path,
) -> Tuple[
    zipfile.ZipFile,
    Optional[Tuple[Dict[str, np.ndarray], Dict[str, np.ndarray]]]
]:
    """If the input object is a TPU program - return that TPU program zip file and None;
    if the input object is a TPU test - return its TPU program and its IO pair of dicts;
    otherwise throw an exception."""

    tpu_program_or_tpu_test_zip = zipfile.ZipFile(path)
    names = set(tpu_program_or_tpu_test_zip.namelist())

    if {_INPUTS_MEMBER_PREFIX + '/', _OUTPUTS_MEMBER_PREFIX + '/', _TPU_PROGRAM_MEMBER_NAME} <= names:
        # TPU test provided!

        # ['inputs/', 'outputs/', 'program.tpu', 'outputs/output_Softmax.bin', 'inputs/input_input.bin']
        tpu_test = tpu_program_or_tpu_test_zip
        tpu_program = zipfile.ZipFile(io.BytesIO(tpu_test.read(_TPU_PROGRAM_MEMBER_NAME)))

        #  Get names of input and output tensors:
        meta_data = _read_meta_data(tpu_program)

        # Read inputs and outputs:
        inputs = dict(_read_tensor_dict(tpu_test, meta_data, _INPUTS_MEMBER_PREFIX))
        outputs = dict(_read_tensor_dict(tpu_test, meta_data, _OUTPUTS_MEMBER_PREFIX))

        return tpu_program, (inputs, outputs)

    elif 'metadata.json' in names:
        # TPU program only is provided
        return tpu_program_or_tpu_test_zip, None
    else:
        raise ValueError('Unrecognized zip-file content')


def _copy_zip_file_to(zip_file, file_name):
    with tempfile.TemporaryDirectory() as temporary_directory:
        zip_file.extractall(temporary_directory)
        shutil.make_archive(file_name, 'zip', temporary_directory)
        os.rename(file_name + '.zip', file_name)


def eval_deviation(data: np.ndarray, reference: np.ndarray, relative_threshold: float = relative_threshold,
         absolute_threshold: float = absolute_threshold) -> np.ndarray:
    abs_error = np.abs(reference - data)
    relative_error = np.abs(abs_error / reference)

    return np.logical_or(relative_error <= relative_threshold, abs_error <= absolute_threshold)



import os
import threading as thr
import numpy as np
import tempfile
import contextlib
from itertools import cycle
from math import ceil
import concurrent.futures
from collections import deque

import pytpu as tpu
from ._helpres import get_data_path, _inspect_input_zip, _copy_zip_file_to, eval_deviation

from typing import Dict
from typing import Any
from typing import Any
from typing import Dict

import pytest


THREADS = 1


ziptests = [
    'non_raw_test_concat[int8-1-1-19].zip',
    'non_raw_test_concat[int8-1-4-19].zip',
    'raw_test_concat[int8-1-1-32].zip',
    'non_raw_test_concat[int8-1-1-32].zip',
    'raw_test_concat[int8-1-1-19].zip',
    'raw_test_concat[int8-1-4-19].zip'
]


def _check_test_result(
    result: Dict[str, np.ndarray],
    expected: Dict[str, np.ndarray],
):
    result_value = next(iter(result.values()))
    expected_value = next(iter(expected.values()))

    if expected_value.dtype in ['uint8', 'int8']:
        assert np.array_equal(result_value, expected_value) == True
        assert np.max(np.abs(result_value - expected_value)) == 0

    elif expected_value.dtype == 'float32':
        assert eval_deviation(result_value, expected_value).all()


def _sequential(
    tpu_program_path: str,
    inputs: Dict[str, np.ndarray],
    outputs: Dict[str, np.ndarray],
):
    inputs = [inputs for _ in range(100)]
    with contextlib.ExitStack() as tpu_device_stack:
        tpu_devices = [tpu_device_stack.enter_context(tpu.Device.open(i)) for i in tpu.Device.list_devices()]
        with contextlib.ExitStack() as tpu_program_stack:
            tpu_programs = [tpu_program_stack.enter_context(tpu.load(tpu_program_path)) for tpu in tpu_devices]
            with contextlib.ExitStack() as tpu_inference_stack:
                programs_merged_inputs = zip(cycle(tpu_programs), range(len(inputs)))

                for tpu_program, input_index in programs_merged_inputs:
                    inference = tpu_inference_stack.enter_context(tpu_program.inference())
                    input_ = inputs[input_index]

                    result = inference.sync(input_)

                    # for zip test
                    if outputs:
                        _check_test_result(result, outputs)

def _threading(
    tpu_program_path: str,
    inputs: Dict[str, np.ndarray],
    outputs: Dict[str, np.ndarray],
):
    _thread_to_inference = dict()
    _device_to_semaphore = dict()

    def worker_init():
        tpu_program = tpu_program_queque.pop()
        thread = thr.current_thread()
        inference = tpu.Inference(tpu_program)
        _thread_to_inference[thread] = inference
        _device_to_semaphore[str(inference._program._device)] = thr.BoundedSemaphore(value=THREADS)

    def target(input_):
        thread = thr.current_thread()
        inference = _thread_to_inference[thread]
        semaphore = _device_to_semaphore[str(inference._program._device)]

        with semaphore:
            result = inference.sync(input_)

        return result

    def calculate_predictions(inputs_lst):
        lst = []
        with concurrent.futures.ThreadPoolExecutor(max_workers=THREADS, initializer=worker_init, initargs=()) as executor:
            futures = []

            for input_ in inputs_lst:
                futures.append(executor.submit(target, input_))

            for future in concurrent.futures.as_completed(futures):
                result = future.result()
                lst.append(result)

        for inference in _thread_to_inference.values():
           inference.close()

        return lst

    list_devices = tpu.Device.list_devices()
    with contextlib.ExitStack() as tpu_device_stack:
        tpu_devices = [tpu_device_stack.enter_context(tpu.Device.open(i)) for i in list_devices]

        with contextlib.ExitStack() as tpu_program_stack:
            tpu_programs = [tpu_program_stack.enter_context(tpu.load(tpu_program_path)) for tpu in tpu_devices]
            tpu_program_queque = deque(tuple(tpu_programs) * ceil(THREADS / len(tpu_devices)))

            # for zip test
            if outputs:
                inputs_lst = [inputs for _ in range(100)]
                predictions = calculate_predictions(inputs_lst)
                for result in predictions:
                    _check_test_result(result, outputs)


@pytest.mark.parametrize("ziptest", ziptests)
def test_ziptest_seq(ziptest: str):

    program_path = os.path.join(get_data_path(), 'ziptests', ziptest)

    tpu_program_zip, inputs_outputs = _inspect_input_zip(program_path)

    # if test zip
    if inputs_outputs:
        inputs, outputs = inputs_outputs
        with tempfile.TemporaryDirectory() as temporary_directory:
            # Dump to a temporary directory:
            program_path = os.path.join(temporary_directory, 'temporary_tpu_program.tpu')
            _copy_zip_file_to(tpu_program_zip, program_path)
            _sequential(program_path, inputs, outputs)


@pytest.mark.parametrize("ziptest", ziptests)
def test_ziptest_thread(ziptest):

    program_path = os.path.join(get_data_path(), 'ziptests', ziptest)

    tpu_program_zip, inputs_outputs = _inspect_input_zip(program_path)

    # if test zip
    if inputs_outputs:
        inputs, outputs = inputs_outputs
        with tempfile.TemporaryDirectory() as temporary_directory:
            # Dump to a temporary directory:
            program_path = os.path.join(temporary_directory, 'temporary_tpu_program.tpu')
            _copy_zip_file_to(tpu_program_zip, program_path)
            _threading(program_path, inputs, outputs)



import contextlib
import os.path
import numpy as np
import pytest

import pytpu as tpu

from pytpu import InferenceError
from pytpu_tests.unit._helpres import get_data_path


def test_HW_params_mismatch():

    with contextlib.ExitStack() as tpu_device_stack:
        tpu_devices = [tpu_device_stack.enter_context(tpu.Device.open(i)) for i in tpu.Device.list_devices()]
        tpu_device = tpu_devices[0]
        dev_info = tpu_device.info()

        # Choose wrong program path
        ddr_word_length = dev_info["axi_word_len"] // 8
        if ddr_word_length == 64:
            program_path = os.path.join(get_data_path(), 'ddr_32/cache_128', '128x128_program.tpu')
        elif ddr_word_length == 32:
            program_path = os.path.join(get_data_path(), 'ddr_64/cache_64', '64x64_fpga_program.tpu')

        with pytest.raises(ValueError): # Catch an exception
            tpu_program = tpu_device_stack.enter_context(tpu_device.load(program_path))
            inference = tpu_device_stack.enter_context(tpu_program.inference())

            shape = (1, 224, 224, 3)
            a = (10 * np.random.normal(size=shape)).astype(np.int8)
            b = (10 * np.random.normal(size=shape)).astype(np.int8)

            input_tensor = {'InputLayer0': a,
                            'InputLayer1': b}

            inference.sync(input_tensor)



"""This is a basic example that demonstrates multithreading TPU inference with pytpu."""
import collections
import contextlib
import threading
import concurrent.futures
from typing import Dict
from typing import Deque
from datetime import datetime
from ._helpres import get_data_path

import numpy as np
import pytpu as tpu
import os

# Number of Python threads bound for each available TPU device
# (None or negative stands for single thread and single TPU)
WORKER_THREADS_PER_TPU_DEVICE = 8   # TODO: getting ....XXX.... here <---------------------------------------------------------------------------- !!!!!!!

# WORKER_THREADS_PER_TPU_DEVICE = None  # Works fine

# PATH_TO_TPU_PROGRAM = 'non_raw_add_1_17_17_128.tpu'
TPU_PROGRAM_NAME = '_1_224_224_3_program.tpu'
# PATH_TO_TPU_PROGRAM = 'pytpu.2/pytpu_tests/data/parallel/'


# NUMBER_OF_INPUT_SAMPLES = int(10 * 1e3)  # number of trials


NUMBER_OF_INPUT_SAMPLES = 10000  # number of trials

_INPUT_IMAGES = [None for i in range(NUMBER_OF_INPUT_SAMPLES)]  # Prepare dummy inputs


def test_parallel():
# if __name__ == '__main__':

    # Each Python thread upon initialization will be bound to a particular TPU device via inference descriptor object,
    # the following dictionary keeps track of that relation:
    _thread_to_inference: Dict[threading.Thread, tpu.Inference] = dict()

    # Upon initialization each worker will pop from this queue an interface descriptor:
    _inference_queue: Deque = collections.deque()


    def worker_init():
        """Initialize new Python worker thread"""

        thread = threading.current_thread()  # Get current working thread
        inference = _inference_queue.pop()  # Pop inference object to bound to

        print(f'New thread {thread} bound to inference descriptor {inference}!')

        _thread_to_inference[thread] = inference  # Bound current thread to an inference description and TPU device


    def worker_func(path: str):
        """ Inference function that is executed in a separate thread by each worker.
        :param path: full path to the i-th input image
        """

        # Get inference descriptor of the current thread:
        thread = threading.current_thread()
        inference = _thread_to_inference[thread]

        # Execute pre-processing function in the same thread as inference thus better this better utilizing parallelism
        # in possible numpy-function called in the pre-processing function.
        # TODO: call pre-processing here
        # input_tensor = _pre_processing(path)  # convert input image to input tensor

        shape = (1, 224, 224, 3)
        a = (10 * np.random.normal(size=shape)).astype(np.int8)
        b = (10 * np.random.normal(size=shape)).astype(np.int8)

        # Prepare dummy input:
        input_tensor = {'InputLayer0': a,
                        'InputLayer1': b}

        # Synchronous call of the inference function (called in a separate, current thread):
        output_tensor_dictionary = inference.sync(input_tensor)

        # Execute post-processing function in the same manner:
        # TODO: call post-processing here

        # Prepare dummy output
        score = output_tensor_dictionary['Add0']

        # print(score.flatten()[:10])
        # print(a.flatten()[:10])
        # print(b.flatten()[:10])

        if np.array_equal(score, a + b):
            print('.', end='')
        else:
            print('X', end='')

        return score


    # Prepare inference loop:
    with contextlib.ExitStack() as exit_stack:

        program_path = os.path.join(get_data_path(), 'parallel', TPU_PROGRAM_NAME)
        # Iterate over available TPU devices, open each device, load the same program
        # and make `WORKER_THREADS_PER_TPU_DEVICE` inference descriptors.
        # This will result in `WORKER_THREADS_PER_TPU_DEVICE` times the number of available devices threads.
        for tpu_device_path in tpu.Device.list_devices():
            tpu_device = exit_stack.enter_context(tpu.Device.open(tpu_device_path))  # open n-th TPU device
            tpu_program = exit_stack.enter_context(tpu_device.load(program_path))  # load TPU program

            if not WORKER_THREADS_PER_TPU_DEVICE:
                # If WORKER_THREADS_PER_TPU_DEVICE is None or <= 0 - run a single thread only:
                i = exit_stack.enter_context(tpu_program.inference())  # Make single inference descriptor
                _inference_queue.append(i)  # Append i-th inference to the queue

                break  # Exit TPU loop now
            else:
                # Make `WORKER_THREADS_PER_TPU_DEVICE` inference descriptors:
                for _ in range(WORKER_THREADS_PER_TPU_DEVICE):
                    i = exit_stack.enter_context(tpu_program.inference())  # Make i-th inference
                    _inference_queue.append(i)  # Append i-th inference to the queue

        # With the _inference_queue populated prepare thread pool:
        with concurrent.futures.ThreadPoolExecutor(max_workers=len(_inference_queue), initializer=worker_init,
                                                   initargs=()) as executor:

            futures = [executor.submit(worker_func, path) for path in _INPUT_IMAGES]  # Submit each input image

            begin_at = datetime.now()

            # Run main inference loop:
            counter = 0
            for future in concurrent.futures.as_completed(futures):
                output_data = future.result()

                # TODO: process scores here
                # print('.', end='')

                counter += 1

                if counter % 80 == 0:
                    print()

            print()

            time_delta = datetime.now() - begin_at
            print(f'Processing of {len(_INPUT_IMAGES)} images complete in {time_delta},'
                  f' average FPS is {len(_INPUT_IMAGES) / time_delta.total_seconds()}')

    ...

    # NOTE: exit_stack will release resources in the reversed order thus first closing inference description then
    # unloading TPU program then closing TPU device.
